# Author: Yuehao Bai | University of Michigan | Feb 2021
# -----------------------------------------------------
# Functions used in simulation with many moments inequalities. Implement
#   - Bai, Santos, and Shaikh (2021), which in turn builds on Romano, Shaikh, and Wolf (2014), henceforth "RSW".
#   - Chernozhukov, Chetverikov, and Kato (2019) two-step procedure with empirical bootstrap, henceforth "CCK".
# -----------------------------------------------------
library("Rcpp")
sourceCpp("bootHD.cpp")

genBaseline <-
  function(model,
           A,
           theta = 0,
           n = 100,
           p = 100,
           eps.dist = "N",
           b.null = 0.8,
           gamma.1 = 0.05,
           gamma.2 = 0.1) {
    # ----------------------------------
    # Generate baseline according to
    #     X[ij] = theta * (1{j <= gamma.1 * p} + veps[ij]) - b * 1{gamma.2 * p < j <= p} + veps[ij]
    #     veps = A' eps, Sigma = A'A
    #
    # eps.dist (all variance 1):
    #     "N": N(0, 1)
    #     "t": t(4) / sqrt 2
    #     "U": U[-sqrt 3, sqrt 3]
    #
    # Models:
    #     1: b = 0, Sigma equi --- 1 on diag, rho o.w.
    #     2: b = 0.8, Sigma equi --- 1 on diag, rho o.w.
    #     3: b = 0, Sigma auto --- rho^|j - k|
    #     4: b = 0.8, Sigma auto --- rho^|j - k|
    # ----------------------------------
    
    if (eps.dist == "N") {
      eps <- matrix(rnorm(n * p), n, p)
    } else if (eps.dist == "t") {
      eps <- rt(n, df = 4) / sqrt(2)
    } else if (eps.dist == "U") {
      eps <- runif(n, -sqrt(3), sqrt(3))
    } else {
      stop("distribution of eps isn't valid")
    }
    
    veps <- eps %*% A
    
    if (model %in% c(1, 3)) {
      b <- 0
    } else {
      b <- b.null
    }
    
    mu <-
      c(rep(theta, gamma.1 * p), rep(0, (gamma.2 - gamma.1) * p), rep(-b, (1 - gamma.2) * p))
    x <-
      matrix(rep(mu, n), n, byrow = T) + (theta + 1) * veps
    list(
      x = x,
      veps = veps,
      model = model,
      eps.dist = eps.dist,
      n = n,
      p = p
    )
  }

testsHD <-
  function(data.obs,
           alpha = 0.05,
           beta.rsw = alpha / 10,
           beta.cck = 0.001,
           n.boot = 1000) {
    # ------------------------------------------
    # Generates results for both RSW and CCK.
    # ------------------------------------------
    x <- data.obs$x
    n <- dim(x)[1]
    p <- dim(x)[2]
    s.hat <- sqrt(apply(x, 2, var) * (n - 1) / n)
    mu.hat <- colMeans(x)
    t.vector <- sqrt(n) * mu.hat / s.hat
    
    # rsw: compute multiplier d of first-step confidence region
    # cck: compute first step critical value
    first.boot <- bootHD(x, B = n.boot)
    mu.boot <- first.boot[1:n.boot, ]
    s.boot <- first.boot[(n.boot + 1): (2 * n.boot), ] * sqrt((n - 1) / n)
    numer.mat <- sqrt(n) * (matrix(mu.hat, n.boot, p, byrow = T) - mu.boot)
    first.boot.rsw.mat <- numer.mat / s.boot
    first.boot.rsw2.mat <- numer.mat / matrix(s.hat, n.boot, p, byrow = T)
    first.boot.rsw <- apply(first.boot.rsw.mat, 1, max)
    first.boot.rsw2 <- apply(first.boot.rsw2.mat, 1, max)
    first.boot.cck <- apply(-first.boot.rsw.mat, 1, max)
    first.boot.cck2 <- apply(-first.boot.rsw2.mat, 1, max)
    
    # rsw: intersect resulting confidence region with third quadrant to get `null vector'
    # cck: moment selection
    d.rsw <- quantile(first.boot.rsw, 1 - beta.rsw)
    lambda.star <- pmin(rep(0, p), mu.hat + d.rsw * s.hat / sqrt(n))
    d.rsw2 <- quantile(first.boot.rsw2, 1 - beta.rsw)
    lambda.star2 <-
      pmin(rep(0, p), mu.hat + d.rsw2 * s.hat / sqrt(n))
    d.cck <- quantile(first.boot.cck, 1 - beta.cck)
    j.first <- t.vector > -2 * d.cck
    d.cck2 <- quantile(first.boot.cck2, 1 - beta.cck)
    j.first2 <- t.vector > -2 * d.cck2
    
    # compute bootstrap critical values
    second.boot <- bootHD(x, B = n.boot)
    mu.boot <- second.boot[1:n.boot, ]
    s.boot <- second.boot[(n.boot + 1): (2 * n.boot), ] * sqrt((n - 1) / n)
    mu.center.mat <- mu.boot - matrix(mu.hat, n.boot, p, byrow = T)
    
    test.stat.boot.rsw <- pmax(0, apply(sqrt(n) * (mu.center.mat + matrix(lambda.star, n.boot, p, byrow = T)) / s.boot, 1, max))
    crit.val.rsw <- max(quantile(test.stat.boot.rsw, 1 - alpha + beta.rsw), 0)
    test.stat.boot.rsw2 <- pmax(0, apply(sqrt(n) * (mu.center.mat + matrix(lambda.star2, n.boot, p, byrow = T)) / matrix(s.hat, n.boot, p, byrow = T), 1, max))
    crit.val.rsw2 <- max(quantile(test.stat.boot.rsw2, 1 - alpha + beta.rsw), 0)
    
    if (sum(j.first) == 0) {
      crit.val.cck <- 0
    } else {
      test.stat.boot.cck <- apply(sqrt(n) * mu.center.mat[, j.first] / s.boot[, j.first], 1, max)
      crit.val.cck <- quantile(test.stat.boot.cck, 1 - alpha + 2 * beta.cck)
    }
    
    if (sum(j.first2) == 0) {
      crit.val.cck2 <- 0
    } else {
      test.stat.boot.cck2 <- apply(sqrt(n) * mu.center.mat[, j.first2] / matrix(s.hat[j.first2], n.boot, sum(j.first2), byrow = T), 1, max)
      crit.val.cck2 <- quantile(test.stat.boot.cck2, 1 - alpha + 2 * beta.cck)
    }
    
    # compute the test
    test.stat <- max(t.vector)
    test.stat.vec <-
      c(max(test.stat, 0), max(test.stat, 0), test.stat, test.stat)
    crit.val.vec <-
      c(crit.val.rsw, crit.val.rsw2, crit.val.cck, crit.val.cck2)
    reject <- test.stat.vec > crit.val.vec
  }

testsHDbetas <-
  function(data.obs,
           alpha = 0.05,
           beta.vec,
           n.boot = 1000) {
    # ------------------------------------------
    # Generates results for both RSW and CCK with a vector of values for beta.
    # ------------------------------------------
    x <- data.obs$x
    n <- dim(x)[1]
    p <- dim(x)[2]
    s.hat <- sqrt(apply(x, 2, var) * (n - 1) / n)
    mu.hat <- colMeans(x)
    t.vector <- sqrt(n) * mu.hat / s.hat
    
    n.beta <- length(beta.vec)
    
    first.boot <- bootHD(x, B = n.boot)
    mu.boot <- first.boot[1:n.boot, ]
    s.boot <- first.boot[(n.boot + 1): (2 * n.boot), ] * sqrt((n - 1) / n)
    numer.mat <- sqrt(n) * (matrix(mu.hat, n.boot, p, byrow = T) - mu.boot)
    first.boot.rsw.mat <- numer.mat / s.boot
    first.boot.rsw2.mat <- numer.mat / matrix(s.hat, n.boot, p, byrow = T)
    first.boot.rsw <- apply(first.boot.rsw.mat, 1, max)
    first.boot.rsw2 <- apply(first.boot.rsw2.mat, 1, max)
    first.boot.cck <- apply(-first.boot.rsw.mat, 1, max)
    first.boot.cck2 <- apply(-first.boot.rsw2.mat, 1, max)
    
    # rsw: intersect resulting confidence region with third quadrant to get `null vector'
    # cck: moment selection
    d.rsw <- rep(NA, n.beta)
    lambda.star <- matrix(NA, p, n.beta)
    d.rsw2 <- rep(NA, n.beta)
    lambda.star2 <- matrix(NA, p, n.beta)
    d.cck <- rep(NA, n.beta)
    j.first <- vector(mode = "list", length = n.beta)
    d.cck2 <- rep(NA, n.beta)
    j.first2 <- vector(mode = "list", length = n.beta)
    
    for (i.beta in 1:n.beta) {
      dd <- quantile(first.boot.rsw, 1 - beta.vec[i.beta])
      lambda.star[, i.beta] <- pmin(rep(0, p), mu.hat + dd * s.hat / sqrt(n))
      d.rsw[i.beta] <- dd
      
      dd <- quantile(first.boot.rsw2, 1 - beta.vec[i.beta])
      lambda.star2[, i.beta] <-
        pmin(rep(0, p), mu.hat + dd * s.hat / sqrt(n))
      d.rsw2[i.beta] <- dd
      
      dd <- quantile(first.boot.cck, 1 - beta.vec[i.beta])
      j.first[[i.beta]] <- t.vector > -2 * dd
      d.cck[i.beta] <- dd
      
      dd <- quantile(first.boot.cck2, 1 - beta.vec[i.beta])
      j.first2[[i.beta]] <- t.vector > -2 * dd
      d.cck2[i.beta] <- dd
    }
    
    second.boot <- bootHD(x, B = n.boot)
    mu.boot <- second.boot[1:n.boot, ]
    s.boot <- second.boot[(n.boot + 1): (2 * n.boot), ] * sqrt((n - 1) / n)
    mu.center.mat <- mu.boot - matrix(mu.hat, n.boot, p, byrow = T)
    
    crit.val.rsw <- rep(NA, n.beta)
    crit.val.rsw2 <- rep(NA, n.beta)
    crit.val.cck <- rep(NA, n.beta)
    crit.val.cck2 <- rep(NA, n.beta)
    
    for (i.beta in 1:n.beta) {
      test.stat.boot.rsw <- pmax(0, apply(sqrt(n) * (mu.center.mat + matrix(lambda.star[, i.beta], n.boot, p, byrow = T)) / s.boot, 1, max))
      crit.val.rsw[i.beta] <- max(quantile(test.stat.boot.rsw, 1 - alpha + beta.vec[i.beta]), 0)
      test.stat.boot.rsw2 <- pmax(0, apply(sqrt(n) * (mu.center.mat + matrix(lambda.star2[, i.beta], n.boot, p, byrow = T)) / matrix(s.hat, n.boot, p, byrow = T), 1, max))
      crit.val.rsw2[i.beta] <- max(quantile(test.stat.boot.rsw2, 1 - alpha + beta.vec[i.beta]), 0)
      
      if (sum(j.first[[i.beta]]) == 0) {
        crit.val.cck[i.beta] <- 0
      } else {
        test.stat.boot.cck <- apply(sqrt(n) * mu.center.mat[, j.first[[i.beta]]] / s.boot[, j.first[[i.beta]]], 1, max)
        crit.val.cck[i.beta] <- quantile(test.stat.boot.cck, 1 - alpha + 2 * beta.vec[i.beta])
      }
      
      if (sum(j.first2[[i.beta]]) == 0) {
        crit.val.cck2[i.beta] <- 0
      } else {
        test.stat.boot.cck2 <- apply(sqrt(n) * mu.center.mat[, j.first2[[i.beta]]] / matrix(s.hat[j.first2[[i.beta]]], n.boot, sum(j.first2[[i.beta]]), byrow = T), 1, max)
        crit.val.cck2[i.beta] <- quantile(test.stat.boot.cck2, 1 - alpha + 2 * beta.vec[i.beta])
      }
    }
    
    # compute the test, results in vector form
    test.stat <- max(t.vector)
    test.stat.vec <-
      c(max(test.stat, 0), max(test.stat, 0), test.stat, test.stat)
    crit.val.mat <-
      cbind(crit.val.rsw, crit.val.rsw2, crit.val.cck, crit.val.cck2)
    reject <- matrix(test.stat.vec, n.beta, 4, byrow = T) > crit.val.mat
  }

testRSW <-
  function(data.obs,
           alpha = 0.05,
           beta.rsw = alpha / 10,
           n.boot = 10000) {
    # ------------------------------------------
    # Romano, Shaikh, and Wolf (2014)
    # ------------------------------------------
    x <- data.obs$x
    n <- dim(x)[1]
    p <- dim(x)[2]
    s.hat <- sqrt(diag(var(x)))
    mu.hat <- colMeans(x)
    t.vector <- sqrt(n) * mu.hat / s.hat
    
    # rsw: compute multiplier d of first-step confidence region
    first.boot <- bootHD(x, B = n.boot)
    mu.boot <- first.boot[1:n.boot, ]
    s.boot <- first.boot[(n.boot + 1): (2 * n.boot), ] * sqrt((n - 1) / n)
    numer.mat <- sqrt(n) * (matrix(mu.hat, n.boot, p, byrow = T) - mu.boot)
    first.boot.rsw.mat <- numer.mat / s.boot
    first.boot.rsw <- apply(first.boot.rsw.mat, 1, max)
    
    d.rsw <- quantile(first.boot.rsw, 1 - beta.rsw)
    lambda.star <- pmin(rep(0, p), mu.hat + d.rsw * s.hat / sqrt(n))
    
    second.boot <- bootHD(x, B = n.boot)
    mu.boot <- second.boot[1:n.boot, ]
    s.boot <- second.boot[(n.boot + 1): (2 * n.boot), ] * sqrt((n - 1) / n)
    mu.center.mat <- mu.boot - matrix(mu.hat, n.boot, p, byrow = T)
    
    test.stat.boot.rsw <- pmax(0, apply(sqrt(n) * (mu.center.mat + matrix(lambda.star, n.boot, p, byrow = T)) / s.boot, 1, max))
    crit.val.rsw <- max(quantile(test.stat.boot.rsw, 1 - alpha + beta.rsw), 0)
    
    test.stat <- max(t.vector)
    reject <- test.stat > crit.val.rsw
  }

testCCK <-
  function(data.obs,
           alpha = 0.05,
           beta.cck = 0.001,
           n.boot = 10000) {
    # ------------------------------
    # Chernozhukov et al. (2019) two-step with EB
    # ------------------------------
    x <- data.obs$x
    n <- dim(x)[1]
    p <- dim(x)[2]
    s.hat <- sqrt(diag(var(x)))
    mu.hat <- colMeans(x)
    t.vector <- sqrt(n) * mu.hat / sqrt(diag(var(x)))
    
    # first step selection
    first.boot <- bootHD(x, B = n.boot)
    mu.boot <- first.boot[1:n.boot, ]
    s.boot <- first.boot[(n.boot + 1): (2 * n.boot), ] * sqrt((n - 1) / n)
    numer.mat <- sqrt(n) * (matrix(mu.hat, n.boot, p, byrow = T) - mu.boot)
    first.boot.rsw.mat <- numer.mat / s.boot
    first.boot.cck <- apply(-first.boot.rsw.mat, 1, max)
    
    d.cck <- quantile(first.boot.cck, 1 - beta.cck)
    j.first <- t.vector > -2 * d.cck
    
    second.boot <- bootHD(x, B = n.boot)
    mu.boot <- second.boot[1:n.boot, ]
    s.boot <- second.boot[(n.boot + 1): (2 * n.boot), ] * sqrt((n - 1) / n)
    mu.center.mat <- mu.boot - matrix(mu.hat, n.boot, p, byrow = T)
    
    if (sum(j.first) == 0) {
      crit.val.cck <- 0
    } else {
      test.stat.boot.cck <- apply(sqrt(n) * mu.center.mat[, j.first] / s.boot[, j.first], 1, max)
      crit.val.cck <- quantile(test.stat.boot.cck, 1 - alpha + 2 * beta.cck)
    }
    
    test.stat <- max(t.vector)
    reject <- test.stat > crit.val.cck
  }